/* Ephemeris.h
 * front end object for the Swiss Ephemeris
 *
 * Copyright (C) 1999-2005 by Georg Fleischmann
 * Author:   Georg Fleischmann
 *
 * created:  1999-11-30
 * modified: 2005-10-19
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the vhf Public License as
 * published by vhf interservice GmbH. Among other things, the
 * License requires that the copyright notices and this notice
 * be preserved on all copies.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the vhf Public License for more details.
 *
 * You should have received a copy of the vhf Public License along
 * with this program; see the file LICENSE. If not, write to vhf.
 *
 * vhf interservice GmbH, Im Marxle 3, 72119 Altingen, Germany
 * eMail: info@vhf.de
 * http://www.vhf.de
 */

#ifndef VHF_H_EPHEMERIS
#define VHF_H_EPHEMERIS

#include	"swephexp.h"

/* planet names and intersections for earth projection in order of objects array
 * (see swephexp.h: SE_XXX)
 */
#define OBJECTS	[NSArray arrayWithObjects:@"Sun", @"Moon", @"Mercury", @"Venus", @"Mars", @"Jupiter", @"Saturn", @"Uranus", @"Neptune", @"Pluto", @"Node", @"True Node", @"Apogee", @"True Apogee", @"Earth", @"Chiron", @"Pholus", @"Ceres", @"Pallas", @"Juno", @"Vesta", nil]

#define DIR_NORMAL	0
#define DIR_REVERSE	1

/* 1:1 from SE_XXX */
typedef enum
{
    SUN        = SE_SUN,
    MOON       = SE_MOON,
    MERCURY    = SE_MERCURY,
    VENUS      = SE_VENUS,
    MARS       = SE_MARS,
    JUPITER    = SE_JUPITER,
    SATURN     = SE_SATURN,
    URANUS     = SE_URANUS,
    NEPTUNE    = SE_NEPTUNE,
    PLUTO      = SE_PLUTO,
    NODE       = SE_MEAN_NODE,
    TRUENODE   = SE_TRUE_NODE,
    APOGEE     = SE_MEAN_APOG,
    TRUEAPOGEE = SE_TRUE_NODE,
    CHIRON     = SE_CHIRON,
    PHOLUS     = SE_PHOLUS,
    CERES      = SE_CERES,
    PALLAS     = SE_PALLAS,
    JUNO       = SE_JUNO,
    VESTA      = SE_VESTA,
    AC         = 21,
    MC         = 22
}EphemerisObjects;

#define O_SUN		SE_SUN
#define O_MOON		SE_MOON
#define O_MERCURY	SE_MERCURY
#define O_VENUS		SE_VENUS
#define O_MARS		SE_MARS
#define O_JUPITER	SE_JUPITER
#define O_SATURN	SE_SATURN
#define O_URANUS	SE_URANUS
#define O_NEPTUNE	SE_NEPTUNE
#define O_PLUTO		SE_PLUTO
#define O_NODE		SE_MEAN_NODE
#define O_TRUENODE	SE_TRUE_NODE
#define O_APOGEE	SE_MEAN_APOG
#define O_TRUEAPOGEE	SE_OSCU_APOG
#define O_CHIRON	SE_CHIRON
#define O_PHOLUS	SE_PHOLUS
#define O_CERES		SE_CERES
#define O_PALLAS	SE_PALLAS
#define O_JUNO		SE_JUNO
#define O_VESTA		SE_VESTA
#define O_AC		21
#define O_MC		22

#define O_ARRAYSIZE	23	// enough room to store all objects

/* structure holding the data of each planet
 */
typedef struct
{
    double	lon;		// the zodiac position (ecliptical longitude)
    double	lat;		// Ecliptic latitude
    double	vLon;		// speed in longitude (deg/day), negative = retrograde
    double	vLat;		// speed in latitude (deg/day), negative = retrograde

    double	rec;		// equatorial rectascension
    double	dec;		// equatorial declination (earth latitude)
    double	vRec;		// speed in rectascension
    double	vDec;		// speed in declination

    double	nodeLon;	// ecliptical longitude of ascending node
    double	nodeLat;	// ecliptical latitude of ascending node
    double	nodeVLon;	// speed of ascending node in longitude
    double	nodeVLat;	// speed of ascending node in latitude
    double	dscNodeLon;	// ecliptical longitude of descending node
    double	dscNodeLat;	// ecliptical latitude of descending node
    double	dscNodeVLon;	// speed of descending node in longitude
    double	dscNodeVLat;	// speed of descending node in latitude

    double	perihelionLon;	// ecliptical longitude of perihelion (closest to Sun)
    double	perihelionLat;	// ecliptical latitude of perihelion
    double	perihelionVLon;	// speed of perihelion in longitude
    double	perihelionVLat;	// speed of perihelion in latitude
    double	aphelionLon;	// ecliptical longitude of aphelion (farthest to Sun)
    double	aphelionLat;	// ecliptical latitude of aphelion
    double	aphelionVLon;	// speed of aphelion in longitude
    double	aphelionVLat;	// speed of aphelion in latitude

    //double	perigeeLon;	// ecliptical longitude of pregigee (closest to Earth)
    //double	perigeeLat;	// ecliptical latitude  of pregigee
    //double	perigeeVLon;	// speed of pregigee in longitude
    //double	perigeeVLat;	// speed of pregigee in latitude
    //double	apogeeLon;	// ecliptical longitude of apogee (farthest to Earth)
    //double	apogeeLat;	// ecliptical latitude  of apogee
    //double	apogeeVLon;	// speed of apogee in longitude
    //double	apogeeVLat;	// speed of apogee in latitude
}EphObject;

@interface Ephemeris:NSObject
{
    NSArray	*objectNames;
    EphObject	objects[30];
    BOOL	calculateMeanNodes;
}

- (id)init;
- (void)setEphemerisPath:(NSString*)path;
- (NSArray*)objectNames;
- (void)setMeanNodes:(BOOL)flag;

- (int)indexForObjectName:(NSString*)name;

- (EphObject*)objectsAtUTC:(NSCalendarDate*)utc;
- (EphObject*)objectsAtUTC:(NSCalendarDate*)utc lat:(float)lat lon:(float)lon;
- (EphObject*)objectsAtUTC:(NSCalendarDate*)utc lat:(double)lat lon:(double)lon
                      elev:(double)alt topocentric:(BOOL)topocentric;
- (EphObject*)objectsAtYear:(int)year month:(int)month day:(int)day
                       hour:(int)hour minute:(int)minute second:(int)second
                  longitude:(double)lon latitude:(double)lat elevation:(double)alt;

- (NSMutableDictionary*)objectDictAtUTC:(NSCalendarDate*)utc;
- (NSMutableDictionary*)objectDictAtUTC:(NSCalendarDate*)utc lat:(float)lat lon:(float)lon;
- (NSMutableDictionary*)objectDictAtUTC:(NSCalendarDate*)utc lat:(double)lat lon:(double)lon
                                   elev:(double)alt topocentric:(BOOL)topocentric;
- (NSMutableDictionary*)currentObjectDict;
- (NSMutableDictionary*)currentSpeedLonDict;	// V lon
- (NSMutableDictionary*)currentNodeDict;
- (NSMutableDictionary*)currentDeclinationDict;	// dec
- (NSMutableDictionary*)currentSpeedDecDict;	// V dec

- (double*)houseAtUTC:(NSCalendarDate*)utc latitude:(float)lat longitude:(float)lon houseCount:(int)houseCnt;

@end

#endif	// VHF_H_EPHEMERIS
